from ..mesh.datatypes import *
from ..attributes import face_normals, face_area
import numpy as np
from numpy.random import randint, random, choice

@allowed_mesh_types(SurfaceMesh)
def sample_vertices_from_surface(mesh : SurfaceMesh, n_pts, sample_normals:bool=False) -> PointCloud:
    assert mesh.is_triangular()
    NF = len(mesh.faces)
    areas = face_area(mesh).as_array()
    areas /= np.sum(areas)
    pts = np.zeros((n_pts,3))
    
    faces = choice(NF, size=n_pts, p=areas) # faces on which we take the points (probability weighted by area)
    for i,f in enumerate(faces):
        pA,pB,pC = (mesh.vertices[_v] for _v in mesh.faces[f])
        rA,rB,rC = np.random.random(3)
        pts[i,:] = (rA*pA+rB*pB+rC*pC)/(rA+rB+rC)
    
    pointcloud = PointCloud()
    pointcloud.vertices += list(pts)
    if sample_normals:
        normals = face_normals(mesh) 
        pc_normals = pointcloud.vertices.create_attribute("normals", float, 3, dense=True)
        for i,f in enumerate(faces):
            pc_normals[i] = normals[f]
    return pointcloud


from . import Vec
import numpy as np
    
class BB2D:
    """
    Axis aligned bounding box in 2 dimensions
    """
    def __init__(self, x1:float, y1:float, x2:float, y2:float):
        """
        Args:
            x1 (float): x coordinate of first point 
            y1 (float): y cooordinate of first point
            x2 (float): x coordinate of second point
            y2 (float): y coordinate of second point
        """
        self._x1 = Vec(min(x1,x2), min(y1,y2))
        self._x2 = Vec(max(x1,x2), max(y1,y2))
        self._d = self._x2 - self._x1
    
    @classmethod
    def of_mesh(cls, mesh : "Mesh", padding:float=0.):
        """Computes the bounding box of all vertices of a mesh, projected into the xy-plane.

        Args:
            mesh (Mesh): input mesh. Can be of any type (only the 'vertices' container is accessed)
            padding (float, optional): slack to be added between the mesh and the box. Defaults to 0 for a tight bounding box.

        Returns:
            BB2D: bounding box of the vertices of the mesh
        """
        padding = max(0, padding) # only allow non-negative numbers
        pt_min = np.min(mesh.vertices._data, axis=0)
        pt_max = np.max(mesh.vertices._data, axis=0)
        return BB2D(pt_min[0] - padding, pt_min[1] - padding, pt_max[0] + padding, pt_max[1] + padding)

    @property
    def left(self) -> float:
        """
        Returns:
            float
        """
        return self._x1.x
    @property
    def right(self) -> float:
        """
        Returns:
            float
        """
        return self._x2.x
    @property
    def top(self) -> float:
        """
        Returns:
            float
        """
        return self._x2.y
    @property
    def bottom(self) -> float:
        """
        Returns:
            float
        """
        return self._x1.y
    @property
    def width(self) -> float:
        """
        Returns:
            float
        """
        return self._d.x
    @property
    def height(self) -> float:
        """
        Returns:
            float
        """ 
        return self._d.y
    @property
    def center(self) -> Vec:
        """Coordinates of the center point

        Returns:
            Vec: center point
        """
        return (self._x1 + self._x2)/2

    @staticmethod
    def intersection(b1,b2):
        """Computes the intersection bounding box of two bounding boxes

        Args:
            b1 (BB2D): first bounding box
            b2 (BB2D): second bounding box

        Returns:
            BB2D: a bounding box representing the intersection
        """
        if BB2D.do_intersect(b1,b2):
            if b1.left < b2.left:
                x1,x2 = b2.left, b1.right
            else:
                x1,x2 = b1.left, b2.right
            if b1.bottom < b2.bottom:
                y1,y2 = b2.bottom, b1.top
            else:
                y1,y2 = b1.bottom, b2.top
            return BB2D(x1,y1,x2,y2)
        return None
    
    def __and__(self,other):
        return BB2D.intersection(self,other)

    @staticmethod
    def do_intersect(b1,b2) -> bool:
        """
        Intersection test between two bounding boxes

        Args:
            b1 (BB2D): first bounding box
            b2 (BB2D): second bounding box

        Returns:
            bool: whether the two BB intersect
        """
        return (b1.left <= b2.right and b1.right >= b2.left) and \
                (b1.bottom <= b2.top and b1.top >= b2.bottom)
    
    @staticmethod
    def union(b1,b2):
        """Computes the union bounding box of two bounding boxes

        Args:
            b1 (BB2D): first bounding box
            b2 (BB2D): second bounding box

        Returns:
            BB2D: a bounding box representing the union
        """
        pmin = np.min((b1._x1, b2._x1), axis=0)
        pmax = np.max((b1._x2, b2._x2), axis=0)
        return BB2D(pmin[0], pmin[1], pmax[0], pmax[1])
    
    def __or__(self, other):
        return BB2D.union(self,other)
    
    def contains_point(self,pt:Vec) -> bool:
        """Point - bounding box intersection predicate.
        If the point is on the boundary of the box, the convention is as follows: inclusive if the point touches the min coordinates, exclusive for the max coordinates

        Args:
            pt (Vec): a query position

        Returns:
            bool: whether the point 'pt' is inside the bounding box.
        """
        return self.left <= pt[0] < self.right and self.bottom <= pt[1] < self.top
    
    def is_empty(self)->bool:
        """Tests if the bounding box encloses an empty domain

        Returns:
            bool: whether the bounding box is empty
        """
        return np.any(self._x1 >= self._x2)

class BB3D:
    """
    Axis Aligned Bounding Box
    """
    def __init__(self, *args):
        if len(args)==6:
            x1,y1,z1,x2,y2,z2 = args
        elif len(args)==2:
            x1,y1,z1 = args[0]
            x2,y2,z2 = args[1]
        else:
            raise Exception("BB3D expected either (x1,y1,z1,x2,y2,z2) or ((x1,y1,z1), (x2,y2,z2)) as argument.")
        self._x1 : Vec = Vec(x1,y1,z1)
        self._x2 : Vec = Vec(x2,y2,z2)
        self._d  : Vec = self._x2 - self._x1

    @classmethod
    def of_mesh(cls, mesh : "Mesh", padding:float = 0.):
        """Computes the bounding box of all vertices of a mesh

        Args:
            mesh (Mesh): input mesh. Can be of any type (only the 'vertices' container is accessed)
            padding (float, optional): slack to be added between the mesh and the box. Defaults to 0 for a tight bounding box.

        Returns:
            BB3D: bounding box of the vertices of the mesh
        """
        pad = Vec(padding, padding, padding)
        pt_min = np.min(mesh.vertices._data, axis=0) - pad
        pt_max = np.max(mesh.vertices._data, axis=0) + pad
        return BB3D(pt_min, pt_max)

    @property
    def min_coords(self) -> Vec:
        """Minimum coordinates of any point inside the box
        The box is an axis-aligned hexahedron which opposite corners are min_coords and max_coords

        Returns:
            Vec: minimum coordinates
        """
        return self._x1
    
    @property
    def max_coords(self) -> Vec:
        """Maximum coordinates of any point inside the box. 
        The box is an axis-aligned hexahedron which opposite corners are min_coords and max_coords

        Returns:
            Vec: maximum coordinates
        """
        return self._x2
    
    @property
    def span(self) -> Vec:
        """Three dimensions of the box

        Returns:
            Vec: dimensions of the box
        """
        return self._d

    @property
    def center(self) -> Vec:
        """Coordinates of the center point

        Returns:
            Vec: center point
        """
        return (self._x1 + self._x2)/2

    @staticmethod
    def intersection(b1,b2):
        """Computes the intersection bounding box of two bounding boxes

        Args:
            b1 (BB3D): first bounding box
            b2 (BB3D): second bounding box

        Returns:
            BB3D: a bounding box representing the intersection
        """
        if BB3D.do_intersect(b1,b2):
            pmin = Vec.zeros(3)
            pmax = Vec.zeros(3)
            for i in range(3):
                if b1.min_coords[i] < b2.min_coords[i]:
                    pmin[i], pmax[i] = b2.min_coords[i], b1.max_coords[i]
                else:
                    pmin[i], pmax[i] = b1.min_coords[i], b2.max_coords[i]
            return BB3D(pmin,pmax)
        return None
    
    def __and__(self,other):
        return BB3D.intersection(self,other)

    @staticmethod
    def do_intersect(b1,b2) -> bool:
        """
        Intersection test between two bounding boxes

        Args:
            b1 (BB3D): first bounding box
            b2 (BB3D): second bounding box

        Returns:
            bool: whether the two BB intersect
        """
        predicates = [b1.min_coords[i] <= b2.max_coords[i] and b1.max_coords[i] >= b2.min_coords[i] for i in range(3)]
        return np.all(predicates)
    
    @staticmethod
    def union(b1,b2):
        """Computes the union bounding box of two bounding boxes

        Args:
            b1 (BB3D): first bounding box
            b2 (BB3D): second bounding box

        Returns:
            BB3D: a bounding box representing the union
        """
        pmin = np.min((b1._x1, b2._x1), axis=0)
        pmax = np.max((b1._x2, b2._x2), axis=0)
        return BB3D(pmin, pmax)
    
    def __or__(self, other):
        return BB3D.union(self,other)

    def contains_point(self,pt:Vec) -> bool:
        """Point - bounding box intersection predicate.
        If the point is on the boundary of the box, the convention is as follows: inclusive if the point touches the min coordinates, exclusive for the max coordinates

        Args:
            pt (Vec): a query position

        Returns:
            bool: whether the point 'pt' is inside the bounding box.
        """
        return (self._x1 <= pt).all() and (pt < self._x2).all()
    
    def is_empty(self)->bool:
        """Tests if the bounding box encloses an empty domain

        Returns:
            bool: whether the bounding box is empty
        """
        return np.any(self.min_coords >= self.max_coords)
    
    


import mouette as M
import sys
from time import time


input_mesh = M.mesh.load(sys.argv[1])

t0 = time()
domain = M.geometry.BB2D.of_mesh(input_mesh, padding=0.1)
qt = M.processing.QuadTree(domain)

# for v in input_mesh.vertices:
#     qt.insert_point(v.x, v.y)

for v in input_mesh.boundary_vertices:
    p = input_mesh.vertices[v]
    qt.insert_point(p.x, p.y)

print(f"Done in {time()-t0} s")
qt_mesh = qt.export_as_polyline()
M.mesh.save(input_mesh, "input.mesh")
M.mesh.save(qt_mesh, "quadtree.mesh")


from ...mesh.datatypes import *
from ...mesh.mesh_data import DataContainer, RawMeshData
from ...geometry import Vec, BB2D
import numpy as np

from dataclasses import dataclass
from enum import Enum

# class SpacePartitionTree:
    
#     class Node:
#         def __init__(self, bounding_box: AABB, depth:int = 0):
#             self.BB : AABB = bounding_box
#             self.depth = depth
#             self.id : int = None
#             self.children : list = []
#             self.parent : int = None
        
#         @property
#         def parent(self):
#             if self.id==1: return None
#             return self.id//2

#         @property
#         def children(self):
#             return 2*self.id, 2*self.id+1
        
#         def subdivide(self):
#             return

#     def __init__(self, dim:int, points: DataContainer, domain : AABB = None):
#         self.points = points
#         self.dim = dim
#         self._id_node = 0
#         self._nodes = [None]
#         bb = self.get_bounding_box() if domain is None else domain
#         _ = self.create_node(bb, 0) # root
#         self._where = points.create_attribute("node_id", int, dense=True)

#     @property
#     def root(self):
#         return self._nodes[1]

#     def create_node(self, bb:AABB, depth:int):
#         new_node = SpacePartitionTree.Node(bb, depth=depth)
#         new_node.id = self._id_node
#         self._id_node += 1
#         self._nodes.append(new_node)
#         return self._id_node

#     def get_bounding_box(self, padding:float=0.):
#         xmin,ymin,zmin = np.min(self.points._data, axis=0)
#         xmax,ymax,zmax = np.max(self.points._data, axis=0)
#         if self.dim==2:
#             pmin,pmax,pdim = Vec(xmin,ymin), Vec(xmax,ymax), Vec(padding,padding)
#         else:
#             pmin,pmax,pdim = Vec(xmin,ymin,zmin), Vec(xmax,ymax,zmax), Vec(padding,padding,padding)
#         return AABB(pmin,pmax-pmin+pdim)

#     def append(pt:Vec):
#         return

class QuadTree:
    
    MAX_PT_PER_NODE = 1

    @dataclass
    class Node:
        id : int # id of the node
        depth : int # depth of the node inside the tree
        parent : int  # id of parent node (-1 if root node)
        corners : list # indices of vertex corners of the node (should be 4 elements)
        children : list = None # id of children nodes
        points: set = None # id of contained points

    def __init__(self, domain : BB2D):
        self.points = DataContainer(id="vertices")
        self._where = self.points.create_attribute("id", int)

        ## Build a treemesh as a rect
        self._tree_vert : PointCloud = PointCloud()
        v0 = Vec(domain.left, domain.bottom, 0.)
        v1 = Vec(domain.right, domain.bottom, 0.)
        v2 = Vec(domain.right, domain.top, 0.)
        v3 = Vec(domain.left, domain.top, 0.)
        self._tree_vert.vertices += [v0,v1,v2,v3]
        
        self._id_node = -1
        self._nodes = []
        self._bb = [] # bounding boxes of nodes

        ## Build a root node
        self._create_node(0, -1, [0,1,2,3])

    def _get_new_node_id(self):
        self._id_node += 1
        return self._id_node

    def _create_node(self, depth, parent, verts):
        assert len(verts)==4
        new_id = self._get_new_node_id()
        new_node = QuadTree.Node(new_id, depth, parent, verts)
        new_node.points = []
        new_node.children = []
        vmin,vmax = self.vertex(verts[0]), self.vertex(verts[2])
        self._bb.append(BB2D(vmin.x, vmin.y, vmax.x, vmax.y))
        self._nodes.append(new_node)
        return new_id

    def vertex(self,i):
        return self._tree_vert.vertices[i]

    def _subdivide(self, node_id):
        """
        Subdivision convention
        v3 -- v6 -- v2
        |     |      |
        |     |      |
        v7 -- v8 --  v5
        |     |      |
        |     |      |
        v0 -- v4 -- v1 
        """
        node : QuadTree.Node = self._nodes[node_id]
        if len(node.children)!=0 : return # not a leaf -> cannot subdivide
        nv = len(self._tree_vert.vertices)
        v0,v1,v2,v3 = node.corners
        p0,p1,p2,p3 = (self.vertex(v) for v in node.corners)
        v4,p4 = nv,   (p0+p1)/2
        v5,p5 = nv+1, (p1+p2)/2
        v6,p6 = nv+2, (p2+p3)/2
        v7,p7 = nv+3, (p0+p3)/2
        v8,p8 = nv+4, (p0+p1+p2+p3)/4
        self._tree_vert.vertices += [p4,p5,p6,p7,p8]
        n1,n2,n3,n4 = (
            self._create_node(node.depth+1, node.id, [v0,v4,v8,v7]),
            self._create_node(node.depth+1, node.id, [v4,v1,v5,v8]),
            self._create_node(node.depth+1, node.id, [v8,v5,v2,v6]),
            self._create_node(node.depth+1, node.id, [v7,v8,v6,v3])
        )
        node.children = [n1,n2,n3,n4]
        for p in node.points:
            self._insert_point(self.points[p], p, node.id)
        node.points = []
    
    def insert_point(self, x:float, y:float):
        """Adds a point inside the Quadtree Structure
        """
        pt = Vec(x, y, 0.)
        if not self._bb[0].contains_point(pt): return
        self.points.append(pt)
        self._insert_point(pt, len(self.points)-1, 0)

    def _insert_point(self, pt, pt_id, node_id):
        node : QuadTree.Node = self._nodes[node_id]
        if len(node.children)!=0:
            # not a leaf -> recursive call to children
            for child in node.children:
                self._insert_point(pt, pt_id, child)
        else:
            BB : BB2D = self._bb[node_id]
            if BB.contains_point(pt):
                self._where[pt_id] = node_id
                node.points.append(pt_id)
                if len(node.points)>QuadTree.MAX_PT_PER_NODE:
                    # trigger subdivision
                    self._subdivide(node_id)
    
    def query(self, pt:Vec):
        return
    
    def find_all_points_in(self, box:BB2D) -> list:
        return
    
    def export_as_polyline(self) -> PolyLine:
        # TODO : make a version with correct connectivity
        out = RawMeshData()
        out.vertices = self._tree_vert.vertices
        for node in self._nodes:
            v0,v1,v2,v3 = node.corners
            out.edges += [(v0,v1), (v1,v2), (v2,v3), (v3,v0)]
        return PolyLine(out)


import pytest
import mouette as M

########## iterators ##########

def consecutive_pairs():
    L = [1,2,3,4,5]
    pairs = list(M.utils.consecutive_pairs_pairs(L))
    assert pairs == [(1,2), (2,3), (3,4), (4,5)]

def test_cyclic_pairs():
    L = [1,2,3,4,5]
    pairs = list(M.utils.cyclic_pairs(L))
    assert pairs == [(1,2), (2,3), (3,4), (4,5), (5,1)]

def test_cyclic_pairs_enumerate():
    L = [1,2,3,4,5]
    pairs = list(M.utils.cyclic_pairs_enumerate(L))
    assert pairs == [((0,1),(1,2)), ((1,2),(2,3)), ((2,3),(3,4)), ((3,4),(4,5)), ((4,0),(5,1))]

def test_consecutive_triplets():
    L = [1,2,3,4,5]
    triplets = list(M.utils.consecutive_triplets(L))
    assert triplets == [(1,2,3), (2,3,4), (3,4,5)]

def test_cyclic_triplets():
    L = [1,2,3,4,5]
    triplets = list(M.utils.cyclic_triplets(L))
    assert triplets == [(5,1,2), (1,2,3), (2,3,4), (3,4,5), (4,5,1)]

def test_cyclic_permutations():
    L = [1,2,3,4]
    perms = [x for x in M.utils.cyclic_permutations(L)]
    assert perms == [
        [1,2,3,4],
        [4,1,2,3],
        [3,4,1,2],
        [2,3,4,1],
    ]

def test_cyclic_perm_enumerate():
    L = [1,2,3,4]
    perms = [x for x in M.utils.cyclic_perm_enumerate(L)]
    assert perms == [
        [(0, 1), (1, 2), (2, 3), (3, 4)],
        [(3, 4), (0, 1), (1, 2), (2, 3)],
        [(2, 3), (3, 4), (0, 1), (1, 2)],
        [(1, 2), (2, 3), (3, 4), (0, 1)],
    ]

def test_offset():
    L = [1,2,3,4,5]
    assert M.utils.offset(L,2)==[3,4,5,1,2]

########## misc ##########

def test_filename():
    assert M.utils.get_filename("path/to/test_file.txt") == "test_file"
    assert M.utils.get_filename("path/to/foo.obj") == "foo"
    assert M.utils.get_filename("path/to/mysuperfile.geogram_ascii") == "mysuperfile"

def test_extension():
    assert M.utils.get_extension("path/to/test_file.txt") == "txt"
    assert M.utils.get_extension("path/to/foo.obj") == "obj"
    assert M.utils.get_extension("path/to/mysuperfile.geogram_ascii") == "geogram_ascii"

def test_replace_extension():
    assert M.utils.replace_extension("path/to/file.txt", ".obj") == "path/to/file.obj"
    assert M.utils.replace_extension("path/to/foo.meh", ".txt") == "path/to/foo.txt"

def test_keyify():
    assert M.utils.keyify(3,2) == (2,3)
    assert M.utils.keyify([1,3,2]) == (1,2,3)
    assert M.utils.keyify((4,5)) == (4,5)

def test_replace_in_list():
    assert M.utils.replace_in_list([1,2,3],2,42) == [1,42,3]
    assert M.utils.replace_in_list(["a","b","a","c"], "a", 16) == [16,"b",16,"c"]

def test_logger():
    log = M.utils.Logger("test_logger", True)
    log.log("this is a test")
    assert True

########## Priority queue ##########

def test_priority_queue_empty():
    p = M.utils.PriorityQueue()
    assert p.empty()

def test_priority_queue_push_pop():
    p = M.utils.PriorityQueue()
    p.push("test",0)
    assert p.front.x == "test"
    assert p.pop().x == "test"
    assert p.empty()

def test_priority_pop_from_empty():
    p = M.utils.PriorityQueue()
    try:
        _ = p.get()
        assert False
    except IndexError:
        assert True

def test_priority_sort():
    p = M.utils.PriorityQueue()
    p.push("a",5)
    p.push("b",2)
    elem = p.get()
    assert elem.x == "b"
    assert elem.priority == 2

########## Union Find ##########

def test_UF_create():
    uf = M.utils.UnionFind()
    assert len(uf)==0
    uf = M.utils.UnionFind(range(3))
    assert len(uf)==3

def test_UF_add():
    uf = M.utils.UnionFind()
    for i in range(3):
        uf.add(i)
    assert len(uf)==3
    assert 1 in uf

def test_UF_get_set():
    uf = M.utils.UnionFind([0,1])
    assert uf[0]==0
    assert uf[1]==1
    uf[0] = 3
    assert uf[0]==3
    assert uf[1]==1

def test_UF_find():
    uf = M.utils.UnionFind(range(3))
    assert uf.find(0)==0
    assert uf.find(1)==1
    assert uf.find(2)==2
    try:
        uf.find(3)
        assert False
    except ValueError:
        assert True

def test_UF_union():
    uf = M.utils.UnionFind(range(3))
    assert not uf.connected(0,1)
    uf.union(0,1)
    assert uf.connected(0,1)
    uf.union(3,0) # should automatically add element 3
    assert uf.connected(0,3)
    assert len(uf)==4

def test_UF_component():
    uf = M.utils.UnionFind(range(3))
    uf.union(0,1)
    assert uf.component(0)==uf.component(1)
    assert uf.component(0)=={0,1}
    try:
        uf.component(42)
        assert True
    except ValueError:
        assert True
    assert uf.components() == [[0,1], [2]]

def test_UF_roots():
    uf = M.utils.UnionFind(range(3))
    uf.union(0,1)
    assert 0 in uf.roots() or 1 in uf.roots()
    assert 2 in uf.roots()

def test_UF_component_mapping():
    uf = M.utils.UnionFind(range(3))
    uf.union(0,1)
    m = uf.component_mapping()
    assert len(m[0])==2
    assert len(m[2])==1

def test_UF_repr():
    uf = M.utils.UnionFind(range(3))
    print(uf)
    assert True


from mouette import geometry as geom
import numpy as np

###### BB2D ######

def test_init():
    bb = geom.BB2D(0.,0.,1.,1.)
    assert bb.left == 0.
    assert bb.right == 1.
    assert bb.bottom == 0.
    assert bb.top == 1.

def test_dim():
    bb = geom.BB2D(-1.,-1., 1., 1.)
    assert bb.width == 2.
    assert bb.height == 2.
    assert bb.center.x == 0.
    assert bb.center.y == 0.

def test_contains():
    for _ in range(10):
        pt_max = 1.001 + geom.Vec.random(2)
        bb = geom.BB2D(-0.01, -0.01, pt_max[0], pt_max[1])
        query_pt = geom.Vec.random(2)
        assert bb.contains_point(query_pt)
        assert not bb.contains_point(query_pt + np.array((2.,0.)))


def test_do_intersect():
    b1 = geom.BB2D(0.,0.,1.,1.)
    b2 = geom.BB2D(0.9,0.9,1.,1.2)
    b3 = geom.BB2D(1.1,1.1,2.2,2.2)
    
    assert geom.BB2D.do_intersect(b1,b2)
    assert not geom.BB2D.do_intersect(b1,b3)

def test_intersection1():
    b1 = geom.BB2D(0., 0.,100.,100.)
    b2 = geom.BB2D(-10., -10., 1.,1.)

    bb_int = b1 & b2
    assert bb_int.left == 0.
    assert bb_int.right == 1.
    assert bb_int.bottom == 0.
    assert bb_int.top == 1. 

def test_intersection2():
    b1 = geom.BB2D(0., 0., 1., 1.)
    b2 = geom.BB2D(2., 2., 3., 3.)

    bb_int = b1 & b2
    assert bb_int is None

def test_union1():
    # b1 and b2 disjoint
    b1 = geom.BB2D(0., 0., 8.,1.)
    b2 = geom.BB2D(10., 10., 10., 80.)

    bb_un = b1 | b2
    assert bb_un.left == 0.
    assert bb_un.right == 10.
    assert bb_un.bottom == 0.
    assert bb_un.top == 80. 

def test_union2():
    # b2 inside b1
    b1 = geom.BB2D(0., 0., 10.,10.)
    b2 = geom.BB2D(1., 1., 2., 2.) 

    bb_un = b1 | b2
    assert bb_un.left == 0.
    assert bb_un.right == 10.
    assert bb_un.bottom == 0.
    assert bb_un.top == 10. 

###### BB3D ######

def test_init():
    bb = geom.BB3D(0.,0.,0., 1.,1.,1.)
    assert np.all(bb.min_coords == np.array((0.,0.,0.)))
    assert np.all(bb.max_coords == np.array((1.,1.,1.)))

def test_dim():
    bb = geom.BB3D(-1.,-1.,-1, 1.,1.,1.)
    assert np.all(bb.span == np.array((2.,2.,2.)))
    assert bb.center.x == 0.
    assert bb.center.y == 0.
    assert bb.center.z == 0.

def test_contains():
    for _ in range(10):
        pt_max = 1.001 + geom.Vec.random(3)
        bb = geom.BB3D(geom.Vec(-.01, -.01, -.01), pt_max)
        query_pt = geom.Vec.random(3)
        assert bb.contains_point(query_pt)
        assert not bb.contains_point(query_pt +geom.Vec(2.,0.,0.))

def test_do_intersect():
    b1 = geom.BB3D(0.,0.,0., 1.,1.,1.)
    b2 = geom.BB3D(0.9,0.9,0.9, 1., 1.2, 1.)
    b3 = geom.BB3D(1.1, 1.1, 1.1, 2.2, 2.2, 2.2)
    
    assert geom.BB3D.do_intersect(b1,b2)
    assert not geom.BB3D.do_intersect(b1,b3)

def test_intersection1():
    b1 = geom.BB3D(0., 0., 0., 100., 100., 100.)
    b2 = geom.BB3D(-10., -10., -10, 1., 1., 1.)

    bb_int = b1 & b2
    assert np.all(bb_int.min_coords == geom.Vec(0.,0.,0.))
    assert np.all(bb_int.max_coords == geom.Vec(1.,1.,1.))

def test_intersection2():
    b1 = geom.BB3D(0., 0., 0., 1., 1., 1.)
    b2 = geom.BB3D(2., 2., 2., 3., 3., 3.)
    bb_int = b1 & b2
    assert bb_int is None

def test_union1():
    # b1 and b2 disjoint
    b1 = geom.BB3D(0., 0., 0., 8., 1., 1.)
    b2 = geom.BB3D(10., 10., 10., 10., 80., 10.)

    bb_un = b1 | b2
    assert np.all(bb_un.min_coords == geom.Vec(0.,0.,0.))
    assert np.all(bb_un.max_coords == geom.Vec(10.,80.,10.))
    

def test_union2():
    # b2 inside b1
    b1 = geom.BB3D(0., 0., 0.,  10., 10., 10.)
    b2 = geom.BB3D(1., 1., 1, 2., 2., 2.) 

    bb_un = b1 | b2
    assert np.all(bb_un.min_coords == b1.min_coords)
    assert np.all(bb_un.max_coords == b1.max_coords)


from mouette.processing import parametrization as PARAM
from utils import *
from data import *

########## LSCM ##########

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere()])
def test_LSCM(m):
    lscm = PARAM.LSCM(m, verbose=False, eigen=True, save_on_corners=True, solver_verbose=False)
    lscm.run()
    assert m.face_corners.has_attribute("uv_coords")

    lscm = PARAM.LSCM(m, verbose=False, eigen=True, save_on_corners=False, solver_verbose=False)
    lscm.run()
    assert m.vertices.has_attribute("uv_coords")

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere()])
def test_LSCM_no_eigen(m):
    lscm = PARAM.LSCM(m, verbose=False, eigen=False, save_on_corners=True, solver_verbose=False)
    lscm.run()
    assert m.face_corners.has_attribute("uv_coords")
    
    lscm = PARAM.LSCM(m, verbose=False, eigen=False, save_on_corners=False, solver_verbose=False)
    lscm.run()
    assert m.vertices.has_attribute("uv_coords")

@pytest.mark.parametrize("m", [surf_cube()])
def test_LSCM_not_disk(m):
    try:
        _ = PARAM.LSCM(m,verbose=False, eigen=False, save_on_corners=False)()
        assert False
    except:
        assert True

@pytest.mark.parametrize("m", [surf_half_sphere()])
def test_export_LSCM(m, tmp_path):
    _ = PARAM.LSCM(m, verbose=False, eigen=False, save_on_corners=True, solver_verbose=False)()
    build_test_io(m, tmp_path, "obj", 2)
    build_test_io(m, tmp_path, "geogram_ascii", 2)

########## Tutte Embedding ##########

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere()])
def test_tutte_circle(m):
    tutte = PARAM.TutteEmbedding(m,verbose=False, save_on_corners=True, boundary_mode="circle")
    tutte.run()
    assert m.face_corners.has_attribute("uv_coords")

    tutte = PARAM.TutteEmbedding(m,verbose=False, save_on_corners=False, boundary_mode="circle")
    tutte.run()
    assert m.vertices.has_attribute("uv_coords")

@pytest.mark.parametrize("m", [surf_circle(), surf_spline(), surf_half_sphere()])
def test_tutte_square(m):
    tutte = PARAM.TutteEmbedding(m,verbose=False, save_on_corners=True, boundary_mode="square")
    tutte.run()
    assert m.face_corners.has_attribute("uv_coords")

    tutte = PARAM.TutteEmbedding(m,verbose=False, save_on_corners=False, boundary_mode="square")
    tutte.run()
    assert m.vertices.has_attribute("uv_coords")

@pytest.mark.parametrize("m", [surf_circle()])
def test_tutte_invalid_boundary(m):
    try:
        tutte = PARAM.TutteEmbedding(m,verbose=False, save_on_corners=True, boundary_mode="foo")
        tutte.run()
        assert False
    except M.utils.InvalidArgumentValueError:
        assert True
        
########## Param Distortion ##########

@pytest.mark.parametrize("m", [surf_half_sphere()])
def test_param_distortion(m):
    _ = PARAM.LSCM(m, verbose=False, eigen=False, save_on_corners=True, solver_verbose=False)()
    disto = PARAM.ParamDistortion(m, "uv_coords", save_on_mesh=True, verbose=False)
    disto.run()

    # check mesh attributes
    for attr_name in ("conformal_dist", "scale_dist", "stretch_dist", "shear_dist", "iso_dist", "det"):
        assert m.faces.has_attribute(attr_name)

    # check summary
    for key in ("conformal", "iso", "shear", "scale", "stretch_mean", "stretch_max"):
        assert key in disto.summary

@pytest.mark.parametrize("m", [surf_uv_sphere_quads()])
def test_quad_quality(m):
    disto = PARAM.QuadQuality(m, save_on_mesh=True, verbose=False)
    disto.run()
    for attr_name in ("conformal_dist", "scale_dist", "stretch_dist", "det"):
        assert m.faces.has_attribute(attr_name)

    # check summary
    for key in ("conformal", "det", "scale", "stretch"):
        assert key in disto.summary


import mouette as M
from math import sin,cos, pi
from random import random

def test_principal_angle():
    for _ in range(10):
        angle_rand = 100*random()
        principal_angle = M.utils.maths.principal_angle(angle_rand)
        assert -pi<=principal_angle<=pi
        assert abs(cos(angle_rand) - cos(principal_angle))<1e-6
        assert abs(sin(angle_rand) - sin(principal_angle))<1e-6

def test_solve_quadratic():

    # A = 0
    assert M.utils.maths.solve_quadratic(0.,1.,3.) == [-3.]

    # no roots
    assert len(M.utils.maths.solve_quadratic(1.,1.,1.))==0

    # single root
    for _ in range(3):
        root = 100*random()
        B = -2*root
        C = root*root
        res = M.utils.maths.solve_quadratic(1.,B,C)
        assert len(res)==1
        assert abs(res[0]-root)<1e-6

    # two roots
    for _ in range(3):
        root1 = 100*random()
        root2 = 100*random()
        root1,root2 = min(root1,root2), max(root1,root2)
        B = -root1-root2
        C = root1*root2
        res = M.utils.maths.solve_quadratic(1.,B,C)
        assert len(res)==2
        res1,res2 = min(res), max(res)
        assert abs(res1-root1)<1e-6
        assert abs(res2-root2)<1e-6


import mouette as M
from data.surfaces import *
import numpy as np

@pytest.mark.parametrize("s", [surf_cube()])
def test_translate(s, tmp_path):
    rand_vec = np.random.random(3)
    cube2 = M.transform.translate(s, rand_vec)
    pmin = np.min(cube2.vertices._data, axis=0)
    pmax = np.max(cube2.vertices._data, axis=0)
    assert (pmin == rand_vec).all()
    assert (pmax == rand_vec + M.Vec(1.,1.,1.)).all()


@pytest.mark.parametrize("s", surfaces)
def test_fit_into_cube(s, tmp_path):
    s2 = M.transform.scale(s, 100*np.random.random())
    s2 = M.transform.translate(s2, 100*M.Vec.random(3))
    s2 = M.transform.fit_into_unit_cube(s2)
    pmin = np.min(s2.vertices._data, axis=0)
    pmax = np.max(s2.vertices._data, axis=0)
    assert (pmin >= M.Vec(0.,0.,0.)).all()
    assert (pmax <= M.Vec(1.,1.,1.)).all()
    

